// -*- mode: C++; c-file-style: "bsd"; tab-width: 4 -*-
// jstroke.c - Pilot User interface for JStroke 1.x.
// JStroke 1.x - Japanese Kanji handwriting recognition technology demo.
// Copyright (C) 1997  Robert E. Wells
// http://wellscs.com/pilot
// mailto:robert@wellscs.com
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program (gpl.html); if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Derived from prior work by Todd David Rudick on JavaDict and StrokeDic.
// Makes use of KANJIDIC data from Jim Breen of Monash University.
// Further credit details available at http://wellscs.com/pilot
// See readme.txt, changelo, and gpl.html for more information.
// ---------------------------------------------------------------------------

#include "jstroke.h"

#include "jstrpilot.h"
#include "jstrokerc.h"
#include "bootdata.h"
#include "memowrite.h"

// ---- global data stuff ----------------------------------------------------

Context* gp = NULL;

// ----- PilotMain -----------------------------------------------------------

DWord  PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
	if (cmd == sysAppLaunchCmdNormalLaunch)	{
		StartApplication();
		EventLoop();
		StopApplication();
	}
	return 0;
}

// ----- StartApplication ----------------------------------------------------

void StartApplication(void) {
	CharPtr cpRaw;
	UInt iLen;

	MemoWriteOpen();
	MemoWrite("JStroke Trace\n");

	// Word alignment matters crucially for Context and the list of ScoreItem
	// pointers.  Allocate them before getting into arbitrary CharPtr offsets.
	// CoPilot mostly handles odd addresses, but the Pilot gets a 
	// "Fatal Exception".  It took me an appalling number of hours to figure
	// this one out - it finally banged me on the head.  -rwells, 970721.

	iLen = (sizeof(Context) + 
			diBigBufLen+1 + diLitBufLen+1);

	// In my experience, CoPilot hangs when I return an error from
	// StartApplication and hence from PilotMain.  So I have just
	// made the app as bulletproof as possible - it soldiers on.
	// -rwells, 970711.

	// Allocate tmp working context for application...
	if (!(cpRaw = MemPtrNew(iLen)))
		ErrBox("Not enough memory.");
	else
	{
		MemSet(cpRaw, iLen, 0); // Initialize to zeroes.
		gp = (Context *) cpRaw; cpRaw += sizeof(Context);

		gp->m_cpBuf = cpRaw; cpRaw += diBigBufLen+1;
		gp->m_cpLit = cpRaw; cpRaw += diLitBufLen+1;

		OpenDatabase();
	}
		
	FrmGotoForm(formID_App);
}

// ----- StopApplication -----------------------------------------------------

void StopApplication(void) {
	VoidPtr vp;

	AppCloseForm();				// Just in case...
	CloseDatabase();

	if (gp)	{
		if ((vp = gp->m_pListMem))
			MemPtrFree(vp);
		MemPtrFree(gp);
		gp = NULL;
	}

	MemoWriteClose();
}

// ----- OpenDatabase --------------------------------------------------------

void OpenDatabase(void) {
	char*         cpAppDbName = dcpAppDbName;
	UInt          index;
	VoidHand      RecHandle;
	Ptr           RecPointer;
	char          nullString = '\0';
	DmOpenRef     pAppDb;
	Err           error;
	Boolean       errReported;
	LocalID       dbId;

	if (!gp)
		return;

	// In my experience, CoPilot hangs when I return an error from
	// StartApplication and hence from PilotMain.  So I have just
	// made the app as bulletproof as possible - it soldiers on.
	// -rwells, 970711.

	OpenStrokeDic();

	if ((dbId = DmFindDatabase(diAppDbCardNo, cpAppDbName)))
		pAppDb = DmOpenDatabase(diAppDbCardNo, dbId, dmModeReadWrite);
	else
		pAppDb = NULL;

	// Try to load preferences if the database exists;
	// but if it is the wrong version, delete it and start over fresh.
	if (pAppDb && LoadPreferences(pAppDb) != 0)	{
		(void) DmCloseDatabase(pAppDb);
		(void) DmDeleteDatabase(diAppDbCardNo, dbId);
		pAppDb = NULL;
	}

	// If database doesn't exist, create it fresh...
	if (!pAppDb) {
		if ((error = DmCreateDatabase(diAppDbCardNo, cpAppDbName,
									  dcAppCreatorId, dcAppDbType,
									  false)) != 0)
			ErrBox2(cpAppDbName, ": Can't create database");
		else {
			if ((dbId = DmFindDatabase(diAppDbCardNo, cpAppDbName)))
				pAppDb = DmOpenDatabase(diAppDbCardNo, dbId, dmModeReadWrite);
			if (!pAppDb)
				ErrBox2(cpAppDbName, ": Can't open database");
		}

		if (pAppDb)
			SavePreferences(pAppDb);

		// Make sure we have all our application records created;
		// initialize them to have a single byte empty string, as
		// a good safe default.

		errReported = false;
		for (index = DmNumRecords(pAppDb); index <= dsDbMaxIndex; index++)
		{
			if (!(RecHandle = DmNewRecord(pAppDb, &index, /*size*/ 1 ))) {
				if (!errReported)
					ErrBox2(cpAppDbName, ": Can't create new record");
				errReported = true;
			}
			else {
				RecPointer = MemHandleLock(RecHandle); // Ignore errors...
				(void) DmWrite(RecPointer, 0, &nullString, 1);
				(void) MemPtrUnlock(RecPointer);
				(void) DmReleaseRecord(pAppDb, index, true);
			}
		}
	}

	// Allocate space in database for RawStrokes, and lock it down so we can
	// use it as memory for the duration.  Note that we have to use
	// DmWrite or DmStrCopy to write the memory...

	if ((RecHandle = DmResizeRecord(pAppDb, dsDbRawStrokes,
									sizeof(RawStroke)*diMaxStrokes))){
		if ((RecPointer = MemHandleLock(RecHandle)))
			gp->m_pRawStrokes = (RawStroke *) RecPointer;
		else
			(void) DmReleaseRecord(pAppDb, dsDbRawStrokes, false);
	}
	if (!gp->m_pRawStrokes)
		ErrBox("Can't allocate RawStrokes");
			
	gp->m_pAppDb = pAppDb;		// Remember in global structure...
}

// ----- CloseDatabase -------------------------------------------------------

void CloseDatabase(void) {
	DmOpenRef     pAppDb = gp->m_pAppDb;
	UInt          indexp1;

	if (!gp)
		return;

	if (gp->m_pRawStrokes) {
		(void) MemPtrUnlock(gp->m_pRawStrokes); 
		(void) DmReleaseRecord(pAppDb, dsDbRawStrokes, false);
		gp->m_pRawStrokes = NULL;
	}

	if (pAppDb) {
		// Delete all records except those we explicitly preserve.
		for (indexp1 = DmNumRecords(pAppDb); indexp1 > dsDbFirstDelete;
			 indexp1--)	{
			(void) DmRemoveRecord(pAppDb, indexp1 - 1);
		}
		DmCloseDatabase(pAppDb);
	}
	gp->m_pAppDb = NULL;

	CloseStrokeDic();
}

// ----- OpenStrokeDic --------------------------------------------------------

void OpenStrokeDic(void) {
	char*         cpDicDbName = dcpDicDbName;
	UInt          index;
	VoidHand      RecHandle;
	Ptr           RecPointer;
	char          nullString = '\0';
	DmOpenRef     pDicDb;
	Err           error;
	LocalID       dbId;
	UInt          dbAttrs;

	if (!gp)
		return;

#ifdef FOR_STROKEDIC_CREATE_1
	// First creation pass: delete database if it already exists.
	if ((dbId = DmFindDatabase(diDicDbCardNo, cpDicDbName))) {
		if ((error = DmDeleteDatabase(diDicDbCardNo, dbId))) {
			ErrBox2(cpDicDbName, ": Can't delete database");
		}
	}
	if ((error = CreateStrokeDic()))
		ErrBox2(cpDicDbName, ": Can't create database");
#endif /*FOR_STROKEDIC_CREATE_1*/

	if (!(dbId = DmFindDatabase(diDicDbCardNo, cpDicDbName)))
		ErrBox2(cpDicDbName, ": Can't find required database");

	pDicDb = NULL;

#ifdef FOR_STROKEDIC_CREATE
	if (dbId) {
		if (!(pDicDb = DmOpenDatabase(diDicDbCardNo, dbId, dmModeReadWrite)))
			ErrBox2(cpDicDbName, ": Can't open database for writing");

		if (pDicDb) {
			if ((error = LoadStrokeDic(pDicDb))) {
				ErrBox2(cpDicDbName, ": Can't update database");
			}
		}
	}
#else /*not FOR_STROKEDIC_CREATE*/
	if (dbId) {
		// Clear the backup attribute once we've had a chance to HotSync
		// it to our PC after database loading, to capture it for later
		// simple download via HotSync.  There is no need for it to
		// keep backing up on the user system, although it may be
		// unavoidable the first time.  -rwells, 19970712.

		dbAttrs = 0;
		(void) DmDatabaseInfo(diDicDbCardNo, dbId, NULL, &dbAttrs,
							  NULL, NULL, NULL, NULL, NULL, NULL,
							  NULL, NULL, NULL);
		if (dbAttrs & dmHdrAttrBackup) {
			dbAttrs &= ~dmHdrAttrBackup;
			(void) DmSetDatabaseInfo(diDicDbCardNo, dbId, NULL, &dbAttrs,
									 NULL, NULL, NULL, NULL, NULL, NULL,
									 NULL, NULL, NULL);
		}

		if (!(pDicDb = DmOpenDatabase(diDicDbCardNo, dbId, dmModeReadOnly)))
			ErrBox2(cpDicDbName, ": Can't open database");
	}
#endif /*not FOR_STROKEDIC_CREATE*/

	gp->m_pDicDb = pDicDb;		// Remember in global structure...
}

// ----- CloseStrokeDic -------------------------------------------------------

void CloseStrokeDic(void) {
	if (!gp)
		return;
	if (gp->m_pDicDb)
		DmCloseDatabase(gp->m_pDicDb);
	gp->m_pDicDb = NULL;
}

// ----- AppOpenForm ---------------------------------------------------------

void AppOpenForm(void) {
	FormPtr   form = FrmGetActiveForm();
	FieldAttrType aFieldAttr;

	if (!gp)
		return;

	// As far as I can tell, all the objects in the form are in non-movable
	// memory.  I haven't found any confirmation of this, but there are no
	// functions for getting handles to form objects, only for getting pointers
	// to form objects.  And I don't see any code doing any kind of release
	// on these pointers.  So there appears to be no downside to remembering
	// pointers to interesting controls as we do here.  -rwells, 970712.

	// Remembers the text entry FieldPtr.
	gp->m_pTextField =
		FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID_text));

	if (gp->m_pTextField)
	{
		// Sets the FieldPtr to use a database record handle, 
		// marking it busy in the database.
		FldSetTextHandle(gp->m_pTextField,
						 (Handle) DmGetRecord(gp->m_pAppDb, dsDbText));

		// Tell Field to resize itself as text is entered, as needed.
		FldGetAttributes(gp->m_pTextField, &aFieldAttr);
		aFieldAttr.dynamicSize = 1;
		FldSetAttributes(gp->m_pTextField, &aFieldAttr);
	}

	// Remembers the popup list ListPtr.
	gp->m_pList =
		FrmGetObjectPtr(form, FrmGetObjectIndex(form, listID_kanji));

	// Remembers the popup trigger ControlPtr
	gp->m_pPopupTrigger =
		FrmGetObjectPtr(form, FrmGetObjectIndex(form, poptrigID_kanji));

	FrmDrawForm(form);			// Display form widgets
	AppDrawForm();				// Display custom form decorations
}

// ----- AppCloseForm --------------------------------------------------------

void AppCloseForm(void) {
	if (!gp)
		return;

	if (gp->m_pTextField)
	{
		// Tell the text entry FieldPtr to forget the database record.
		FldSetTextHandle(gp->m_pTextField, NULL); 
		// Release the busy flag on the database record.
		DmReleaseRecord(gp->m_pAppDb, dsDbText, false);// Ignore changes...
		gp->m_pTextField = NULL;
	}
}

// ----- LoadPreferences -----------------------------------------------------

Err LoadPreferences(DmOpenRef pAppDb) {
	Err error = 0;
	Err error2;
	VoidHand hPref;
	Pref *pPref;

	hPref = DmGetRecord(pAppDb, dsDbPref);
	if (!hPref || (MemHandleSize(hPref) != sizeof(Pref)))
		error = 1;
	else {
		pPref = (Pref *) MemHandleLock(hPref);

		if (!pPref || pPref->m_PrefVersion != dsPrefVersion)
			error = 1;
		else {
			// Handle loading new preferences here...
		}
		if ((error2 = MemHandleUnlock(hPref)) != 0)
			error = error2;
	}
	if ((error2 = DmReleaseRecord(pAppDb, dsDbPref, false)) != 0)
		error = error2;

	return error;
}

// ----- SavePreferences -----------------------------------------------------

void SavePreferences(DmOpenRef pAppDb) {
	UInt          index = dsDbPref; // Probably depends on dsDbPref == 0...
	VoidHand      RecHandle;
	Ptr           RecPointer;
	Pref          prefBlk;

	prefBlk.m_PrefVersion = dsPrefVersion;
	// Add new Preference fields here...

	if (DmNumRecords(pAppDb) <= dsDbPref)
		RecHandle = DmNewRecord(pAppDb, &index, sizeof(Pref));
	else
		RecHandle = DmGetRecord(pAppDb, index);

	if (RecHandle)
	{
		if (!(RecPointer = MemHandleLock(RecHandle)))
		{
			(void) DmWrite(RecPointer, 0, &prefBlk, sizeof(Pref));
			(void) MemPtrUnlock(RecPointer); 
		}
	}
	(void) DmReleaseRecord(pAppDb, index, true);  
}

// ----- EventLoop -----------------------------------------------------------

void EventLoop(void)
{
	short err;
	int formID;
	FormPtr form;
	EventType event;

	do {
		EvtGetEvent(&event, 200); // Timeout after 2 seconds...
		
		if (SysHandleEvent(&event))
			continue;
		if (MenuHandleEvent((void *)0, &event, &err))
			continue;

		if (event.eType == frmLoadEvent) {
			formID = event.data.frmLoad.formID;
			form = FrmInitForm(formID);
			FrmSetActiveForm(form);
			switch (formID) {
			case formID_App:
				FrmSetEventHandler(form, (FormEventHandlerPtr)
								   AppEventHandler);
				break;
			}
		}
		FrmDispatchEvent(&event);
	} while(event.eType != appStopEvent);
}

// ----- AppEventHandler ----------------------------------------------------

Boolean AppEventHandler(EventPtr event) {
	int       handled = 0;

	switch (event->eType) {
	case frmOpenEvent:
		AppOpenForm();
		handled = 1;
		break;
    
	case frmCloseEvent:
		AppCloseForm();
		// We leave handle = 0, to get default Form closing behavior.
		break;

	case ctlSelectEvent:  // A control button was pressed and released.
		switch (event->data.ctlEnter.controlID)
		{
		case buttonID_erase:
			AppErase();
			handled = 1;
			break;
		case buttonID_replay:
			AppReplay();
			handled = 1;
			break;
		case buttonID_recognize:
			AppRecognize();
			handled = 1;
			break;
		case buttonID_angles:
			AppAngles();
			handled = 1;
			break;
		}
		break;

	case menuEvent:
		switch (event->data.menu.itemID)
		{
		case menuitemID_about:
			FrmAlert(alertID_about);
			handled = 1;
			break;
		case menuitemID_copy:
			// user has to select (highlight) the text to copy
			FldCopy(gp->m_pTextField);
			handled = 1;
			break;
		case menuitemID_paste:
			FldPaste(gp->m_pTextField);
			handled = 1;
			break;
		}
		break;

	case penDownEvent:
	case penMoveEvent:
	case penUpEvent:
		handled = AppPenEvent(event);
		break;

	case nilEvent:
		break;
	}
	return handled;
}

// ----- AppPenEvent --------------------------------------------------------

Boolean AppPenEvent(EventPtr pEv) {
	Context*  rgp = gp;
	int       handled = 0;
	SWord     ex = pEv->screenX;
	SWord     ey = pEv->screenY;
	int       etype = pEv->eType;
	RawStroke* rsp;
	UInt      iLen;

	//        Use builtin systrap from <UI/Rect.h>.
	Boolean   inBounds = RctPtInRectangle(ex, ey, &Rect_bounds);

	if (!rgp)					// Things are grim if no context...
		return false;

	if (!(rsp = rgp->m_pRawStrokes)) // Grim if no memory to store strokes...
		return false;

	switch (rgp->m_eState)
	{
	case eStarting:
		if (inBounds && etype != penUpEvent)
		{
			rgp->m_sCurStroke = 0;
			rsp = &(rgp->m_pRawStrokes[rgp->m_sCurStroke]);
			iLen = 0;
			if (DmWrite(rgp->m_pRawStrokes,
						(((CharPtr) &(rsp->m_len)) -
						 ((CharPtr) rgp->m_pRawStrokes)),
						&iLen, sizeof(UInt)))
				ErrBox("DmWrite rsp 1");

			rgp->m_bHasSave = false;
			rgp->m_bHasRaw = false;
			rgp->m_iSmoothIn = 0;
			rgp->m_iSmoothOut = 0;
			MemSet(rgp->m_iSmoothRawX, diSmoothCount*sizeof(UInt), 0);
			MemSet(rgp->m_iSmoothRawY, diSmoothCount*sizeof(UInt), 0);
			AppPenSmooth(ex, ey, false);

			rgp->m_bHasRaw = true;
			rgp->m_sRawX = ex;
			rgp->m_sRawY = ey;

			handled = 1;
			rgp->m_eState = eStroking;
		}
		break;

	case eStroking:
		if (inBounds)
		{
			AppPenSmooth(ex, ey, (etype == penUpEvent));
			rgp->m_bHasRaw = true;
			rgp->m_sRawX = ex;
			rgp->m_sRawY = ey;

			handled = 1;
		}
		if (etype == penUpEvent)
		{
			if (!inBounds && rgp->m_bHasRaw) {
				AppPenIntercept(rgp->m_sRawX, rgp->m_sRawY, &ex, &ey);
				AppPenSmooth(ex, ey, true);
			}
			rgp->m_eState = eStrokeDone;
			rgp->m_bHasRaw = false;
		}
		break;

	case eStrokeDone:
		if (inBounds && etype != penUpEvent &&
			rgp->m_sCurStroke < diMaxStrokes)
		{
			rgp->m_sCurStroke++;
			rsp = &(rgp->m_pRawStrokes[rgp->m_sCurStroke]);

			iLen = 0;
			if (DmWrite(rgp->m_pRawStrokes,
						(((CharPtr) &(rsp->m_len)) -
						 ((CharPtr) rgp->m_pRawStrokes)),
						&iLen, sizeof(UInt)))
				ErrBox("DmWrite rsp 2");

			rgp->m_bHasSave = false;
			rgp->m_bHasRaw = false;
			rgp->m_iSmoothIn = 0;
			rgp->m_iSmoothOut = 0;
			MemSet(rgp->m_iSmoothRawX, diSmoothCount*sizeof(UInt), 0);
			MemSet(rgp->m_iSmoothRawY, diSmoothCount*sizeof(UInt), 0);

			AppPenSmooth(ex, ey, false);

			rgp->m_bHasRaw = true;
			rgp->m_sRawX = ex;
			rgp->m_sRawY = ey;

			handled = 1;
			rgp->m_eState = eStroking;
		}
		break;
	}

	return handled;
}

// ----- AppPenIntercept -----------------------------------------------------

void AppPenIntercept(SWord iRawX, SWord iRawY, SWord* ipX, SWord* ipY) {
	SWord iX = *ipX;
	SWord iY = *ipY;
	SWord iDiff, iDelta;

	// MemoWrite2d("*** rx=", iRawX);
	// MemoWrite2d(" ry=", iRawY);
	// MemoWrite2d(" ix=", iX);
	// MemoWrite2d(" iy=", iY);
	// MemoWrite("\n");

	// We know that ipX,ipY is out of bounds, but we have to case it out...
	// We know that iRawX,iRawY is in bounds, which helps.

	if (iX < Rect_x) {
		iDelta   = (iDiff = (iY - iRawY));
		iDelta  *= (iDiff = (iRawX - Rect_x));
		iDelta <<= 1;
		iDelta  /= (iDiff = (iRawX - iX));
		iDelta = ((iDelta + 1) >> 1);
		iY = iRawY + iDelta;
		iX = Rect_x;
	}
	else if (iX > Rect_xt) {
		iDelta   = (iDiff = (iY - iRawY));
		iDelta  *= (iDiff = (Rect_xt - iRawX));
		iDelta <<= 1;
		iDelta  /= (iDiff = (iX - iRawX));
		iDelta = ((iDelta + 1) >> 1);
		iY = iRawY + iDelta;
		iX = Rect_xt;
	}

	if (iY < Rect_y) {
		iDelta   = (iDiff = (iX - iRawX));
		iDelta  *= (iDiff = (iRawY - Rect_y));
		iDelta <<= 1;
		iDelta  /= (iDiff = (iRawY - iY));
		iDelta = ((iDelta + 1) >> 1);
		iX = iRawX + iDelta;
		iY = Rect_y;
	}
	else if (iY > Rect_yt) {
		iDelta   = (iDiff = (iX - iRawX));
		iDelta  *= (iDiff = (Rect_yt - iRawY));
		iDelta <<= 1;
		iDelta  /= (iDiff = (iY - iRawY));
		iDelta = ((iDelta + 1) >> 1);
		iX = iRawX + iDelta;
		iY = Rect_yt;
	}

	*ipX = iX;
	*ipY = iY;
}

// ----- AppPenSmooth --------------------------------------------------------

void AppPenSmooth(SWord ex, SWord ey, Boolean bStrokeDone) {
	Context* rgp = gp;
	UInt iIn, iOut, iModIn, iOutDelta, iOutMax;

	iIn = rgp->m_iSmoothIn++;
	iModIn = iIn % diSmoothCount;
	rgp->m_iSmoothRawX[iModIn] = ex;
	rgp->m_iSmoothRawY[iModIn] = ey;
	iIn++;

	iOut = rgp->m_iSmoothOut;
	do {
		if ((iOutDelta = iOut) > diSmoothMaxDelta)
			iOutDelta = diSmoothMaxDelta;
		iOutMax = iOut + iOutDelta;

		if (iOutMax >= iIn)
			break;

		AppPenSmoothOut(iOut-iOutDelta, iOutMax);
		iOut++;
	} while (true);
	rgp->m_iSmoothOut = iOut;

	if (bStrokeDone) {
		for ( ; iOut < iIn; iOut++) {
			iOutDelta = iIn - iOut - 1;
			AppPenSmoothOut(iOut-iOutDelta, iOut+iOutDelta);
		}
	}
}

// ----- AppPenSmoothOut -----------------------------------------------------

void AppPenSmoothOut(UInt iOutMin, UInt iOutMax) {
	Context* rgp = gp;
	UInt iX, iY, iOut, iOutMod;
	UInt iCount = (iOutMax - iOutMin + 1);
	iX = iY = 0;

	for (iOut = iOutMin; iOut <= iOutMax; iOut++) {
		iOutMod = iOut % diSmoothCount;
		iX += rgp->m_iSmoothRawX[iOutMod];
		iY += rgp->m_iSmoothRawY[iOutMod];
	}

	// Do rounded average of th set of points.
	iX = ((((iX << 1) / iCount) + 1) >> 1);
	iY = ((((iY << 1) / iCount) + 1) >> 1);

	AppPenSave(iX, iY);
}

// ----- AppPenSave ----------------------------------------------------------

void AppPenSave(SWord ex, SWord ey) {
	Context*  rgp = gp;

	if (AppPenThin(ex, ey)) {
		if (rgp->m_bHasSave)
			AppPenDraw(rgp->m_sSaveX, rgp->m_sSaveY, ex, ey);
		else
			rgp->m_bHasSave = true;

		rgp->m_sSaveX = ex;
		rgp->m_sSaveY = ey;
	}
}

// ----- AppPenThin ----------------------------------------------------------

Boolean AppPenThin(SWord ex, SWord ey) {
	Context*   rgp = gp;
	RawStroke* rsp = &(rgp->m_pRawStrokes[rgp->m_sCurStroke]);
	UInt       iLen = rsp->m_len;
	Long       iAng0, iAng1, iAng2, iAngT, iAngDif1, iAngDif2, iAngDifT;
	Long       iAbsAngDif1, iAbsAngDif2, iAbsAngDifT;
	Long       xdif0, ydif0, xdif1, ydif1, xdif2, ydif2, x1,x2,y1,y2;
	Long       xdifT, ydifT;

	if (iLen >= 2) {
		xdif1 =  (rsp->m_x[iLen-1] - rsp->m_x[iLen-2]);
		ydif1 = -(rsp->m_y[iLen-1] - rsp->m_y[iLen-2]);

		xdif2 =  (ex - rsp->m_x[iLen-1]);
		ydif2 = -(ey - rsp->m_y[iLen-1]);

		iAng1 = Angle32(xdif1, ydif1);
		iAng2 = Angle32(xdif2, ydif2);

		// This one is very harmless.  We discard the mid-point of 3,
		// if the first angle and the second angle are identical,
		// or if the second point has not moved from the mid-point.

		if (iAng1 == iAng2 || iAng2 == 32) {
			iLen--;
			rgp->m_iThinCnt--;
		}
		else if (iLen >= 3) {
			xdif0 =  (rsp->m_x[iLen-2] - rsp->m_x[iLen-3]);
			ydif0 = -(rsp->m_y[iLen-2] - rsp->m_y[iLen-3]);
			iAng0 = Angle32(xdif0, ydif0);

			iAngDif1 = iAng1 - iAng0;
			iAngDif2 = iAng2 - iAng1;

			if ((iAbsAngDif1 = iAngDif1) < 0)
				iAbsAngDif1 = -iAngDif1;
			if ((iAbsAngDif2 = iAngDif2) < 0)
				iAbsAngDif2 = -iAngDif2;

			// This is one step trickier.  We discard the two mid-points
			// of four, if the second and third difference in angle are
			// opposite in sign, and the angle doesn't change more than
			// 2/32 (e.g. 11.25 degrees, or 3:45 minutes).

			if (iAbsAngDif1 <= 2 && iAbsAngDif2 <= 2 &&
				((iAngDif1 < 0) != (iAngDif2))) {
				iLen -= 2;
				rgp->m_iThinCnt -= 2;
			}
			else {
				xdifT =  (ex - rsp->m_x[iLen-2]);
				ydifT = -(ey - rsp->m_y[iLen-2]);
				iAngT = Angle32(xdifT, ydifT);
				iAngDifT = iAngT - iAng0;
				if ((iAbsAngDifT = iAngDifT) < 0)
					iAbsAngDifT = -iAngDifT;
				
				// Here we turn gentle bends into rougher stuff.
				if (iAbsAngDifT <= 3) {
					iLen -= 2;
					rgp->m_iThinCnt -= 2;
				}
			}
		}
	}

	if (iLen < diMaxXyPairs) {
		AppPenStore(ex, ey, iLen);
		rgp->m_iRawCnt++;
		rgp->m_iThinCnt++;
		return true;
	}
	else
		return false;
}

// ----- AppPenStore ---------------------------------------------------------

void AppPenStore(SWord ex, SWord ey, UInt idx) {
	Context*  rgp = gp;
	RawStroke* rsp = &(rgp->m_pRawStrokes[rgp->m_sCurStroke]);
	Byte bVal;
	UInt iLen = idx + 1;

	if (idx < diMaxXyPairs && iLen <= diMaxXyPairs) {
		bVal = (Byte) ex;
		if (DmWrite(rgp->m_pRawStrokes,
					(((CharPtr) &(rsp->m_x[idx])) -
					 ((CharPtr) rgp->m_pRawStrokes)),
					&bVal, sizeof(Byte)))
			ErrBox("DmWrite rsp 3x");

		bVal = (Byte) ey;
		if (DmWrite(rgp->m_pRawStrokes,
					(((CharPtr) &(rsp->m_y[idx])) -
					 ((CharPtr) rgp->m_pRawStrokes)),
					&bVal, sizeof(Byte)))
			ErrBox("DmWrite rsp 3y");

		if (DmWrite(rgp->m_pRawStrokes,
					(((CharPtr) &(rsp->m_len)) -
					 ((CharPtr) rgp->m_pRawStrokes)),
					&iLen, sizeof(UInt)))
			ErrBox("DmWrite rsp 3 len");
	}
}

// ----- AppPenDraw ----------------------------------------------------------
// Broaden the line into more of an italic pen...
// Thanks to Roger E Critchlow Jr., San Francisco, California,
// for sharing his doodle code which suggested how to provide
// different pen effects easily.  I don't use his code, which
// used general pen descriptions, but rather hard code his
// fine italic pen pattern, which seemed most appropriate to
// drawing Japanese characters.  -rwells, 970726.

void AppPenDraw(SWord x1, SWord y1, SWord x2, SWord y2) {
	WinDrawLine(x1-1, y1,   x2-1, y2);
	WinDrawLine(x1,   y1,   x2,   y2);
	WinDrawLine(x1,   y1-1, x2,   y2-1);
	WinDrawLine(x1+1, y1-1, x2+1, y2-1);
}

// ----- AppErase ------------------------------------------------------------

void AppErase(void) {
	Word iLen;

	gp->m_eState = eStarting;
	gp->m_iRawCnt = 0;
	gp->m_iThinCnt = 0;
	WinEraseRectangle(&Rect_bounds, 0);
	AppListReplace(AppEmptyList());

	iLen = FldGetTextLength(gp->m_pTextField);
	FldDelete(gp->m_pTextField, 0, iLen);
}

// ----- AppReplay -----------------------------------------------------------

void AppReplay(void) {
	Context*      rgp = gp;
	UInt          iStrokeCnt, iStroke, iLen, idx;
	RawStroke*    rsp;
	Byte*         bpX;
	Byte*         bpY;
	RectangleType Rect_pnt;

	WinEraseRectangle(&Rect_bounds, 0);
	iStrokeCnt = rgp->m_sCurStroke + 1;

	// Draw the character using the normal italic pen.
	for (iStroke = 0; iStroke < iStrokeCnt; iStroke++) {
		rsp = &(rgp->m_pRawStrokes[iStroke]);
		iLen = rsp->m_len;
		bpX = rsp->m_x;
		bpY = rsp->m_y;
		
		for (idx = 1; idx < iLen; idx++)
			AppPenDraw(bpX[idx-1], bpY[idx-1], bpX[idx], bpY[idx]);
	}

	// Now erase just the control point pixels...
	for (iStroke = 0; iStroke < iStrokeCnt; iStroke++) {
		rsp = &(rgp->m_pRawStrokes[iStroke]);
		iLen = rsp->m_len;
		bpX = rsp->m_x;
		bpY = rsp->m_y;
		
		for (idx = 0; idx < iLen; idx++) {
			RctSetRectangle(&Rect_pnt, bpX[idx], bpY[idx], 1, 1);
			WinEraseRectangle(&Rect_pnt, 0);
		}
	}
}

// ----- AppDrawForm ---------------------------------------------------------

void AppDrawForm(void) {
	WinDrawRectangle(&Rect_top, 0);
	WinDrawRectangle(&Rect_left, 0);
	WinDrawRectangle(&Rect_right, 0);
	WinDrawRectangle(&Rect_bot, 0);
}

// ----- AppRecognize ---------------------------------------------------------

void AppRecognize(void) {
	Context*      rgp = gp;
	UInt          iStrokeCnt;
	DmOpenRef     pDicDb;
	VoidHand      RecHandle;
	CharPtr       cpStrokeDic;
	ListMem*      pListMem;
	ListPtr       pList;
	StrokeScorer *pScorer;
	ControlPtr    pPopupTrigger;
	VoidPtr       vp;
	UInt          iListCount;
	UInt          iListHeight;
	UInt          iListLeft;
	UInt          iListTop;
	ULong         iBeginTick, iElapsedTicks;
	Word          iLen;
	CharPtr       cp;

	AppReplay();				// Repaint simplified character.

	cp = "Converting...";
	FldInsert(gp->m_pTextField, cp, StrLen(cp));
	FldRecalculateField(gp->m_pTextField, true);

	iBeginTick = TimGetTicks();

	if (!rgp) {
		return;
	}
	
	switch (rgp->m_eState) {
	case eStarting:
		ErrBox("First draw a Kanji character to recognize");
		return;
	case eStroking:				// Can't happen...
		ErrBox("ERROR: eStroking");
		return;
	case eStrokeDone:
		break;
	}

	iStrokeCnt = rgp->m_sCurStroke + 1;

	if (!(pDicDb = rgp->m_pDicDb)) {
		ErrBox("Sorry, JStrokeDicDB not found");
		return;
	}

	if (!(RecHandle = DmGetRecord(pDicDb, iStrokeCnt))) {
		ErrBox("DmGetRecord err");
		return;
	}

	if (!(cpStrokeDic = (CharPtr) MemHandleLock(RecHandle))) {
		(void) DmReleaseRecord(pDicDb, iStrokeCnt, false);
		ErrBox("MemHandleLock err");
		return;
	}
		
	// Create a StrokeScorer to do the real work...
	// Later pass it an array of actual stroke summary metrics.
	// Later arrange for it to be an "object" that can be "run"
	// repeatedly from the idle event to work on the evaluation
	// in the background during user pause time...  -rwells, 970712.
	// ( Half done - OWT 930904 )

	pScorer = StrokeScorerCreate (cpStrokeDic, 
								  rgp->m_pRawStrokes,
								  iStrokeCnt);

	if (pScorer)
	{
		// Process some database definitions : maxCnt == -1 => all
		StrokeScorerProcess(pScorer, -1);
		// Get the results
		pListMem = StrokeScorerTopPicks(pScorer);
		StrokeScorerDestroy(pScorer);

		// Make the list of top picks the new list contents...
		AppListReplace(pListMem);
	}

	if (MemPtrUnlock(cpStrokeDic))
		ErrBox("MemPtrUnlock err");
	if (DmReleaseRecord(pDicDb, iStrokeCnt, false))
		ErrBox("DmReleaseRecord err");

	// Report the conversion elapsed time...

	iElapsedTicks = TimGetTicks() - iBeginTick;
	cp = gp->m_cpBuf;
	StrIToA(cp, (iElapsedTicks / 100));
	cp += StrLen(cp);
	*cp++ = '.';
	*cp++ = '0' + (iElapsedTicks / 10) % 10;
	*cp++ = '0' + (iElapsedTicks) % 10;
	StrCopy(cp, " Seconds.");
	
	iLen = FldGetTextLength(gp->m_pTextField);
	FldDelete(gp->m_pTextField, 0, iLen);
	FldInsert(gp->m_pTextField, gp->m_cpBuf, StrLen(gp->m_cpBuf));
	FldRecalculateField(gp->m_pTextField, true);

#ifdef FOR_MEMOWRITE	
	MemoWrite2d("s", iStrokeCnt);
	MemoWrite("  ");
	if (pListMem && pListMem->m_argc > 0)
		MemoWriteLen(pListMem->m_argv[0], 2); // top SJIS pick.
	MemoWrite2("  ", gp->m_cpBuf);
	MemoWrite2d("  ", gp->m_iThinCnt);
	MemoWrite2d("/", gp->m_iRawCnt);
	MemoWrite("\n");
#endif /*FOR_MEMOWRITE*/
}

// ----- AppAngles -----------------------------------------------------------

#define diAppAnglesBufMax 80

void AppAngles(void) {
    int sp;
	VoidPtr       spVoid;
	Context*      rgp = gp;
	UInt          idx;
	ULong         off;
	RawStroke*    rsp;
	UInt          iAng32;
	Long          xdif, ydif, x1,x2,y1,y2;

	if (!rgp || !rgp->m_pRawStrokes) {
		ErrBox("ERROR: no gp or no raw strokes");
		return;
	}

	switch (rgp->m_eState) {
	case eStarting:
		ErrBox("First draw a Kanji stroke, then check its path angle");
		return;
	case eStroking:				// Can't happen...
		ErrBox("ERROR: eStroking");
		return;
	case eStrokeDone:
		break;
	}

	rsp = &(rgp->m_pRawStrokes[rgp->m_sCurStroke]);

	x2 = rsp->m_x[rsp->m_len-1];
	x1 = rsp->m_x[0];
	y2 = rsp->m_y[rsp->m_len-1];
	y1 = rsp->m_y[0];

	xdif = x2 - x1;
	ydif = y1 - y2; // Reversed since screen y origin is at top...

	iAng32 = Angle32(xdif, ydif);

	ErrBox2("Angle is ", StrIToA(rgp->m_cpLit, iAng32));

}

// ----- AppListReplace ------------------------------------------------------

void AppListReplace(ListMem* pListMem) {
	Context*      rgp;
	ListPtr       pList;
	ControlPtr    pPopupTrigger;
	UInt          iListCount;
	UInt          iListLeft;
	UInt          iListTop;
	UInt          iListHeight;
	VoidPtr       vp;

	if (!pListMem ||
		!(pListMem->m_argc > 0) ||
		!(rgp = gp) || 
		!(pList = rgp->m_pList) ||
		!(pPopupTrigger = rgp->m_pPopupTrigger))
		return;

	// Limit list count to maximum number we believe are displayable
	// without scrolling...

	iListCount = pListMem->m_argc;
	if (iListCount > diMaxListCount)
		iListCount = diMaxListCount;

	// Calculate the top left position for the popup list,
	// and adjust the length of the list to match its contents.
	// We would like to pop it up on top of the popup trigger,
	// but we will shift it up the screen to accommadate more
	// choices if necessary up to the height of the screen.

	iListLeft = pPopupTrigger->bounds.topLeft.x + diPopupXOffset;
	iListTop = pPopupTrigger->bounds.topLeft.y;
	iListHeight = iListCount * diFontLineHeight;

	while (iListTop >= diFontLineHeight &&
		   iListTop + iListHeight > diScreenHeight) {
		iListTop -= diFontLineHeight;
	}
	if (iListTop + iListHeight > diScreenHeight) {
		iListTop = 0;
	}

	LstSetListChoices(pList, pListMem->m_argv, iListCount);
	LstSetPosition(pList, iListLeft, iListTop);
	LstSetHeight(pList, iListCount);
	LstSetSelection(pList, 0); // Pre-select top choice.
				
	// This makes the choice visible on the form,
	// and triggers the redisplay.
	CtlSetLabel(pPopupTrigger, pListMem->m_argv[0]);

	if ((vp = rgp->m_pListMem))
		MemPtrFree(vp);
	rgp->m_pListMem = pListMem;
}
// ----- end of jstroke.c ----------------------------------------------------
